// Diffie Hellman Functions

#include "types.h"
#include "large_math.h"
#include "diffie_helman.h"
#include "fast_modulo.h"
#include "random.h"
#include "primes.h"
#include "serialize.h"
#include "hash_func.h"
#include "signature.h"
#include "xtea.h"
#include <string.h>

// Define this to reduce encryption overhead
// (supported in versions above 1.7)
//
#define USE_SMALL_ENCRYPT_DECRYPT_PACKET_FUNCTIONS


#ifdef USE_SMALL_ENCRYPT_DECRYPT_PACKET_FUNCTIONS
static const u32 encrypt_overhead  = small_encrypt_packet_overhead_bytes;
#else
static const u32 encrypt_overhead  = packet_overhead_bytes;
#endif


// Check g, g^x, p for validity
//
static int check_gxp(const u32 *g, const u32 *x, const u32 *p, const u32 len_x32)
{
if(less_then_two(g, len_x32)) return 0;

if(less_then_two(x, len_x32)) return 0;

if(!(p[len_x32-1]&0x80000000)) return 0; // Full size

if(0x03 != (p[0]&0x03)) return 0; // Two lsbs must be 0x03

if(multi_compare(p, g, len_x32) <= 0) return 0; // g must be less then p

if(multi_compare(p, x, len_x32) <= 0) return 0; // X must be less then p

if(!multi_compare(g, x, len_x32)) return 0; // g != x

// We are not checking g and x for being generators modulo P
// We are not checking P for being safe prime
// because this would be too slow

return 1;
}

//
// Convert DH shared secret (u32[128] into encryption/decryption keys, u32[18] each 
//
static void dh_secret_to_keys(COLLAPSE *collapse, u32 *k0, u32 *k1, u32 *dh_secret)
{
u32 hash[2*encrypt_decrypt_key_size_x32];

// Hash

collapse->hash(hash, dh_secret, dh_length_x32*sizeof(u32));
 
// Fill 

copy(k0, hash, encrypt_decrypt_key_size_x32);
copy(k1, hash + encrypt_decrypt_key_size_x32, encrypt_decrypt_key_size_x32);

// Cleanup

collapse->reset();
cheap_random_number(dh_secret, dh_length_x32); 
cheap_random_number(hash, 2*encrypt_decrypt_key_size_x32);
reset_stack();
}


TX_SECURE_SESSION::TX_SECURE_SESSION(u32 length_x32)
{
len = length_x32;
x = new u32[len];
p = new u32[len];
g = new u32[len];
gxp = new u32[len];
gyp = new u32[len];
signature = new u32[len];
far_signature_decoded = new u32[len];
common_secret = new u32[len];
tmp_data = new u8[sizeof(SERTIFICATE) + sizeof(SIGNATURE)];
far_hash = new u32[len];

test_generator_modulo_p = new TEST_GENERATOR_MODULO_P(len);
mpm = new MONTGOMERY_POW_MOD(len);
digital_signature = new DIGITAL_SIGNATURE(len);

ds = new SIGNATURE;
far_signature = new SIGNATURE;
far_certificate = new SERTIFICATE;

collapse = new COLLAPSE(2*encrypt_decrypt_key_size_x32);
dh_hash_1 = new DH_HASH(len);
initialized_session = new INITIALIZED_SESSION;

session_is_initialized = 0;
}

TX_SECURE_SESSION::~TX_SECURE_SESSION()
{
TX_SECURE_SESSION::reset();    

delete initialized_session;
delete dh_hash_1;
delete collapse;
delete far_certificate;
delete far_signature;
delete ds;    
delete digital_signature;    
delete mpm;
delete test_generator_modulo_p;

delete[] far_hash;
delete[] tmp_data;
delete[] common_secret;
delete[] far_signature_decoded;
delete[] signature;
delete[] gyp;
delete[] gxp;
delete[] g;
delete[] p;
delete[] x;
}


//-----------------
// Initiate session; returns number of bytes
// 

u32 TX_SECURE_SESSION::init_session_1(u8 *open_data)
{
session_is_initialized = 0;    

// Select P for session

get_p(p);

// Generate G for session
// g must be less then P
// g should be primitive root modulo P

do {
   
   // Force g length one bit smaller then P

   do {
      random_number(g, len);
      g[len-1] &= 0x7FFFFFFF;
      g[len-1] |= 0x40000000;
      } while(less_then_two(g, len)); // just for sanity
 
   } while(!test_generator_modulo_p->test_generator_modulo_p(g, p)); 

// Generate DH secret for session  

random_number(x, len);

// Generate open data X = (g^x) mod p

mpm->multi_pow_mod(gxp, g, x, p);


u32 size_bytes = sizeof(u32)*len;

memcpy(open_data, g, size_bytes);
u32 offset = size_bytes;

memcpy(open_data + offset, p, size_bytes);
offset += size_bytes;

memcpy(open_data + offset, gxp, size_bytes);
offset += size_bytes;

return offset;
}

//-----------------

u32 TX_SECURE_SESSION::process_response_confirm_session_1(u8 *confirmation_data, u8 const *response, // Confirmation data, response from B to A
                         SECRET_DATA const* secret_data, u8 const *password, const u32 password_length_bytes, 
                         int (*verify_certificate)(SERTIFICATE const*))
{
session_is_initialized = 0;    

u32 len_bytes = sizeof(u32)*len;

memcpy(gyp, response, len_bytes);
u32 offset = len_bytes;

// Initialize session data

if(!check_gxp(g, gyp, p, len)) // invalid Y ? 
 {
 TX_SECURE_SESSION::reset();
 return 0;
 }

mpm->multi_pow_mod(common_secret, gyp, x, p);
cheap_random_number(x, len);

// Convert shared secret to keys

dh_secret_to_keys(collapse, encryption_key, decryption_key, common_secret);
cheap_random_number(common_secret, len);

// Init session

initialized_session->init(NULL, encryption_key, decryption_key);

cheap_random_number(encryption_key, encrypt_decrypt_key_size_x32);
cheap_random_number(decryption_key, encrypt_decrypt_key_size_x32);

// Decrypt certificate + signature


if((sizeof(SERTIFICATE) + sizeof(SIGNATURE)) != initialized_session->decrypt(tmp_data, response + offset, 
                                                              sizeof(SERTIFICATE) + sizeof(SIGNATURE) + encrypt_overhead))
 {
 TX_SECURE_SESSION::reset();
 return 0; // Decryption failed
 }

memcpy(far_certificate, tmp_data, sizeof(SERTIFICATE));
memcpy(far_signature, tmp_data + sizeof(SERTIFICATE), sizeof(SIGNATURE));
cheap_random_number(tmp_data, sizeof(SERTIFICATE) + sizeof(SIGNATURE));

// Validate certificate

if(!digital_signature->certificate_looks_ok(far_certificate)) 
 {
 TX_SECURE_SESSION::reset();
 return 0; // some problem with certificate
 }

// Compute hash of g,p,gxp, gyp

dh_hash_1->hash(signature, g, p, gxp, gyp);

signature[len-1] &= 0x3FFFFFFF; // truncate to 4094 bits

// Remember into session

initialized_session->session_fingerprint(signature); // 256 bits (u32[8]) from low side 

// Decode signature of the other side

digital_signature->decode_signature(far_hash ,far_signature, far_certificate);

if(memcmp(far_hash, signature, len*sizeof(u32)))
 {
 TX_SECURE_SESSION::reset();
 return 0;  // signature doesn't match DH parameters
 }

// Validate certificate externally

if(NULL != verify_certificate)
 {
 if(!verify_certificate(far_certificate))
  { 
  TX_SECURE_SESSION::reset();
  return 0;
  }
 }

// Here everything must be OK
// Prepare response

if(!digital_signature->encode_signature(ds, signature, secret_data, password, password_length_bytes))
 {
 TX_SECURE_SESSION::reset();
 return 0; 
 }

memcpy(tmp_data, &secret_data->certificate, sizeof(SERTIFICATE));
memcpy(tmp_data + sizeof(SERTIFICATE), ds, sizeof(SIGNATURE));
 
u32 ret_value = initialized_session->encrypt(confirmation_data, tmp_data, sizeof(SERTIFICATE) + sizeof(SIGNATURE));

if(!ret_value)
 {
 TX_SECURE_SESSION::reset();
 return 0;  // just in case
 }

cheap_random_number(tmp_data, sizeof(SERTIFICATE) + sizeof(SIGNATURE));
ds->reset();

session_is_initialized = 1;
initialized_session->session_id(far_certificate->fingerprint);

return ret_value;
}


int TX_SECURE_SESSION::session_is_active_1(u32 *session_id)
{
if(session_is_initialized)
 {
 if(session_id != NULL) 
  {
  copy(session_id,  far_certificate->fingerprint, fingerprint_length_x32);   
  }
 }

return session_is_initialized;
}

//----------

int TX_SECURE_SESSION::start_initialized_session(INITIALIZED_SESSION *is)
{
    
if(session_is_initialized) 
 {
 *is = *initialized_session; 
 }
else 
 {
 is->reset();   
 }

return session_is_initialized;
}

//-------------------------------

void TX_SECURE_SESSION::reset()
{

cheap_random_number(x, len);
cheap_random_number(p, len);
cheap_random_number(g, len);
cheap_random_number(gxp, len);
cheap_random_number(gyp, len);
cheap_random_number(common_secret, len);
cheap_random_number(signature, len);
cheap_random_number(far_signature_decoded, len);
cheap_random_number(far_hash, len);
cheap_random_number(tmp_data, sizeof(SERTIFICATE) + sizeof(SIGNATURE));



ds->reset();
test_generator_modulo_p->reset();
mpm->reset();

digital_signature->reset();
collapse->reset();
initialized_session->reset();

far_certificate->reset();
far_signature->reset();

//-------
session_is_initialized = 0;
cheap_random_number(encryption_key, encrypt_decrypt_key_size_x32);
cheap_random_number(decryption_key, encrypt_decrypt_key_size_x32);
reset_stack();
}
//
// Operator /=/ to copy initialized session
// This is shallow copy
// Only for processing multiple responses
// 
TX_SECURE_SESSION & TX_SECURE_SESSION::operator = (const TX_SECURE_SESSION &t)
{
if(this != &t) 
 {
 copy(x, t.x, len);    
 copy(p, t.p, len);
 copy(g, t.g, len);
 copy(gxp, t.gxp, len);
 copy(signature, t.signature, len);
 copy(far_signature_decoded, t.far_signature_decoded, len);

 session_is_initialized = t.session_is_initialized;
 copy(encryption_key, t.encryption_key, encrypt_decrypt_key_size_x32);
 copy(decryption_key, t.decryption_key, encrypt_decrypt_key_size_x32);
 }

return *this;
}

u32 TX_SECURE_SESSION::get_user_info_1(u8* user_info)
{
if(!session_is_initialized) return 0;

if(user_info != NULL)
 {
 copy(user_info,far_certificate->user_info, far_certificate->user_info_length_bytes);
 }

return far_certificate->user_info_length_bytes;
}


//=====================================


RX_SECURE_SESSION::RX_SECURE_SESSION(u32 length_x32)
{
len = length_x32;
g  = new u32[len];
p  = new u32[len];
y  = new u32[len];
gxp = new u32[len];
gyp = new u32[len];
signature = new u32[len];
far_signature_decoded = new u32[len];
common_secret = new u32[len];
tmp_response_data = new u8[sizeof(SERTIFICATE) + sizeof(SIGNATURE)];
far_hash = new u32[len];

digital_signature = new DIGITAL_SIGNATURE(len);
ds = new SIGNATURE;
far_certificate = new SERTIFICATE;
mpm = new MONTGOMERY_POW_MOD(len);
collapse = new COLLAPSE(2*encrypt_decrypt_key_size_x32);
dh_hash_1 = new DH_HASH(len);
initialized_session = new INITIALIZED_SESSION;
session_is_initialized = 0;
}

RX_SECURE_SESSION::~RX_SECURE_SESSION()
{
RX_SECURE_SESSION::reset();

delete initialized_session;
delete dh_hash_1;
delete collapse;    
delete mpm;
delete far_certificate;
delete ds;
delete digital_signature;

delete[] far_hash;
delete[] tmp_response_data;
delete[] common_secret;
delete[] far_signature_decoded;
delete[] signature;
delete[] gyp;
delete[] gxp;
delete[] y;
delete[] p;
delete[] g;
}

void RX_SECURE_SESSION::reset()
{
cheap_random_number(g, len);
cheap_random_number(p, len);
cheap_random_number(y, len);
cheap_random_number(gxp, len);
cheap_random_number(gyp, len);
cheap_random_number(signature, len);
cheap_random_number(far_signature_decoded, len);
cheap_random_number(common_secret, len);
cheap_random_number(tmp_response_data, sizeof(SERTIFICATE)+sizeof(SIGNATURE));
cheap_random_number(far_hash, len);


initialized_session->reset();
digital_signature->reset();
ds->reset();
far_certificate->reset();
mpm->reset();
collapse->reset();
dh_hash_1->reset();
session_is_initialized = 0;
cheap_random_number(encryption_key, encrypt_decrypt_key_size_x32);
cheap_random_number(decryption_key, encrypt_decrypt_key_size_x32);
reset_stack();
}



int RX_SECURE_SESSION::session_is_active_1(u32 *session_id)
{

if(session_is_initialized)
 {
 if(session_id != NULL) 
  {
  copy(session_id,  far_certificate->fingerprint, fingerprint_length_x32);   
  }
 }

return session_is_initialized;
}


int RX_SECURE_SESSION::start_initialized_session(INITIALIZED_SESSION *is)
{
    
if(session_is_initialized) 
 {
 *is = *initialized_session;
 }
else 
 {
 is->reset();   
 }
 
return session_is_initialized;
}



u32 RX_SECURE_SESSION::response_to_init_session_1(u8 *response_data, u8 const *open_data, SECRET_DATA const* secret_data, 
                                                 u8 const *password, const u32 password_length_bytes)
{
session_is_initialized = 0;    

// Deserialize 

u32 size_bytes = len*sizeof(u32);

memcpy(g, open_data, size_bytes);
u32 offset = size_bytes;

memcpy(p, open_data + offset, size_bytes);
offset += size_bytes;

memcpy(gxp, open_data + offset, size_bytes);

if(!check_gxp(g, gxp, p, len))
 {
 RX_SECURE_SESSION::reset();
 return 0; //    
 }


// Generate secret y

random_number(y, len);

// Calculate (g^y) mod P

mpm->multi_pow_mod(gyp, g, y, p); // (g^y) mod p

// Calculate ((g^x)^y) mod P = common secret

mpm->multi_pow_mod(common_secret, gxp, y, p); // ((g^x)^y) mod P

// Destroy Y 

cheap_random_number(y, len);

// Compute session keys

dh_secret_to_keys(collapse, decryption_key, encryption_key, common_secret);

// Destroy secret

cheap_random_number(common_secret, len);

// Initialize encryption/decryption

initialized_session->init(NULL, encryption_key, decryption_key);   

// Destroy keys

cheap_random_number(encryption_key, encrypt_decrypt_key_size_x32);
cheap_random_number(decryption_key, encrypt_decrypt_key_size_x32);

// Prepare data

memcpy(tmp_response_data, &secret_data->certificate, sizeof(SERTIFICATE));
u32 tmp_offset = sizeof(SERTIFICATE);

dh_hash_1->hash(signature, g, p, gxp, gyp);

// truncate to 4094 bits

signature[len-1] &= 0x3FFFFFFF;

if(!digital_signature->encode_signature(ds, signature, secret_data, password, password_length_bytes))
 {
 reset();
 return 0; // wrong password
 }

memcpy(tmp_response_data + tmp_offset, ds, sizeof(SIGNATURE));
tmp_offset += sizeof(SIGNATURE);

// Serialize and encrypt 

memcpy(response_data, gyp, size_bytes);
offset = size_bytes;


offset += initialized_session->encrypt(response_data + offset, tmp_response_data, tmp_offset);


// Destroy unnecessary data (note: keep hash value for verification)

cheap_random_number(tmp_response_data, tmp_offset);
ds->reset();
digital_signature->reset();
dh_hash_1->reset();

return offset;
}

int RX_SECURE_SESSION::check_confirmation_1(u8 const *confirmation, int (*verify_certificate)(SERTIFICATE const*))
{
// Decrypt certificate + signature

if((sizeof(SERTIFICATE) + sizeof(SIGNATURE)) != initialized_session->decrypt(tmp_response_data, confirmation, 
                                                              sizeof(SERTIFICATE) + sizeof(SIGNATURE) + encrypt_overhead))
 {
 RX_SECURE_SESSION::reset();
 return 0; // Decryption failed
 }

memcpy(far_certificate, tmp_response_data, sizeof(SERTIFICATE));
memcpy(ds, tmp_response_data + sizeof(SERTIFICATE), sizeof(SIGNATURE));
cheap_random_number(tmp_response_data, sizeof(SERTIFICATE) + sizeof(SIGNATURE));

// Validate certificate

if(!digital_signature->certificate_looks_ok(far_certificate)) 
 {
 RX_SECURE_SESSION::reset();
 return 0; // some problem with certificate
 }

// Decode signature of the other side

digital_signature->decode_signature(far_hash , ds, far_certificate);

if(memcmp(far_hash, signature, len*sizeof(u32)))
 {
 RX_SECURE_SESSION::reset();
 return 0;  // signature doesn't match DH parameters
 }

// Validate certificate externally

if(NULL != verify_certificate)
 {
 if(!verify_certificate(far_certificate))
  { 
  RX_SECURE_SESSION::reset();
  return 0;
  }
 }

// Here everything must be OK

session_is_initialized = 1;
initialized_session->session_id(far_certificate->fingerprint);
initialized_session->session_fingerprint(signature);

return 1;
}


u32 RX_SECURE_SESSION::get_user_info_1(u8* user_info)
{
if(!session_is_initialized) return 0;

if(user_info != NULL)
 {
 copy(user_info,far_certificate->user_info,far_certificate->user_info_length_bytes);
 }

return far_certificate->user_info_length_bytes;
}


//===================================

void INITIALIZED_SESSION::init(u32 const *session_id, u32 const *encrypt_key, u32 const* decrypt_key)   
{
if(session_id != NULL) copy(id, session_id, fingerprint_length_x32);

encryptor.init(encrypt_key);
decryptor.init(decrypt_key);
 
initialized = 1;
}

INITIALIZED_SESSION::INITIALIZED_SESSION()
{
INITIALIZED_SESSION::reset();    
}

INITIALIZED_SESSION::~INITIALIZED_SESSION()
{
INITIALIZED_SESSION::reset();    
}

void INITIALIZED_SESSION::reset()
{
initialized = 0;    
cheap_random_number(id, fingerprint_length_x32);
cheap_random_number(dh_fingerprint, fingerprint_length_x32);
encryptor.reset();
decryptor.reset();
reset_stack();
}

u32 * INITIALIZED_SESSION::session_id()
{
return (initialized) ?  id : NULL;
}

void INITIALIZED_SESSION::session_id(u32 *ptr)
{
if(ptr == NULL)
 {
 initialized = 0;
 }
else
 {
 initialized = 1;
 copy(id, ptr, fingerprint_length_x32); 
 }
}

u32 * INITIALIZED_SESSION::session_fingerprint()
{
return (initialized) ?  dh_fingerprint : NULL;
}

void INITIALIZED_SESSION::session_fingerprint(u32 *ptr)
{
if(ptr == NULL)
 {
 initialized = 0;
 }
else
 {
 initialized = 1;
 copy(dh_fingerprint, ptr, fingerprint_length_x32); 
 }
}



u32 INITIALIZED_SESSION::encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes)  // Packet encryption, returns length of encrypted block (= src_length + 16)
{
#ifdef USE_SMALL_ENCRYPT_DECRYPT_PACKET_FUNCTIONS
return (initialized) ? encryptor.small_encrypt(dest, src, src_len_bytes) : 0;
#else
return (initialized) ? encryptor.encrypt(dest, src, src_len_bytes) : 0;
#endif
}

u32 INITIALIZED_SESSION::decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes) // length           // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)
{
#ifdef USE_SMALL_ENCRYPT_DECRYPT_PACKET_FUNCTIONS
return (initialized) ? decryptor.small_decrypt(dest, src, src_len_bytes) : 0;
#else
return (initialized) ? decryptor.decrypt(dest, src, src_len_bytes) : 0;
#endif
}

u64 INITIALIZED_SESSION::total_rx_bytes()
{
return decryptor.total_rx_bytes();
}

u64 INITIALIZED_SESSION::dropped_rx_bytes()
{
return decryptor.dropped_rx_bytes();
}

u64 INITIALIZED_SESSION::invalid_packets()
{
return decryptor.invalid_packets();
}

u64 INITIALIZED_SESSION::late_packets()
{
return decryptor.late_packets();
}



void SECRET_DATA::reset()
{
encoded_secret_key.reset();
certificate.reset();
data_is_initialized = 0;
reset_stack();
}

//-----------------------------------------------------------------------------------

EL_GAMAL_MODE::EL_GAMAL_MODE()
{
y = new u32[signature_length_x32];
common_secret = new u32[signature_length_x32];
open_key = new u32[signature_length_x32];
mpm = new MONTGOMERY_POW_MOD(signature_length_x32);
ds = new DIGITAL_SIGNATURE(signature_length_x32);
collapse = new COLLAPSE(2*encrypt_decrypt_key_size_x32);
}

EL_GAMAL_MODE::~EL_GAMAL_MODE()
{
EL_GAMAL_MODE::reset();    
delete collapse;
delete ds;
delete mpm;
delete[] open_key;
delete[] common_secret;
delete[] y;    
}

void EL_GAMAL_MODE::reset()
{
cheap_random_number(open_key, signature_length_x32);
cheap_random_number(y, signature_length_x32);
cheap_random_number(common_secret, signature_length_x32);

cheap_random_number(data_hash, 8);
cheap_random_number(encryption_key, encrypt_decrypt_key_size_x32);


signature.reset();
keystream.reset();
mpm->reset();
ds->reset();
collapse->reset();
reset_stack();
}



void EL_GAMAL_MODE::prepare_encryption_data(SERTIFICATE const *sert)
{
// Generate secret Y    

random_number(y, signature_length_x32);

// Generate (g^y) mod P

mpm->multi_pow_mod(open_key, sert->g, y, sert->p);

// Generate common secret ((g^x)^y) mod P

mpm->multi_pow_mod(common_secret, sert->y, y, sert->p);

// Generate encryption key

u32 dummy[encrypt_decrypt_key_size_x32];

dh_secret_to_keys(collapse, encryption_key, dummy, common_secret);

cheap_random_number(dummy, encrypt_decrypt_key_size_x32);
cheap_random_number(y, signature_length_x32); // cleanup
cheap_random_number(common_secret, signature_length_x32); // cleanup
reset_stack();
}

void EL_GAMAL_MODE::prepare_decryption_data(SECRET_DATA const* secret_data, u8 const * password, const u32 password_length_bytes)
{

// Decode key

decode_secret_key(y, &secret_data->encoded_secret_key, password, password_length_bytes);

// Generate common secret ((g^x)^y) mod P

mpm->multi_pow_mod(common_secret, open_key, y, secret_data->certificate.p);


u32 dummy[encrypt_decrypt_key_size_x32];

dh_secret_to_keys(collapse, encryption_key, dummy, common_secret);


cheap_random_number(dummy, encrypt_decrypt_key_size_x32);
cheap_random_number(y, signature_length_x32); // cleanup
cheap_random_number(common_secret, signature_length_x32); // cleanup
reset_stack();
}

//-------------------------------------------------------------


u64 EL_GAMAL_MODE::encrypt(FILE *outfile, FILE *infile, 
                SERTIFICATE const *sert,  // destination  
                SECRET_DATA const *secret_data,  // signature 
                u8 const *password, const u32 password_length_bytes)
{

int signature_included = 0;

if((NULL != secret_data) || (NULL != password) || (0 != password_length_bytes))
 {
 if((NULL == secret_data) ||  (NULL == password) || (0 == password_length_bytes)) 
  {
  reset();
  return 0;
  }

 if(!check_password(&secret_data->encoded_secret_key, password, password_length_bytes)) 
  {
  reset();
  return 0;
  }
 signature_included = 1;
 }

fseek(infile, 0, SEEK_END);
u64 data_len = ftell(infile);
fseek(infile, 0, SEEK_SET);
fseek(outfile, 0, SEEK_SET);

prepare_encryption_data(sert);

fwrite(sert->fingerprint, sizeof(u32), fingerprint_length_x32, outfile);
fputc(signature_included, outfile);
fwrite(&data_len, sizeof(u64), 1, outfile);
fwrite(open_key, sizeof(u32), signature_length_x32, outfile); 

keystream.init_key(encryption_key);

zero(data_hash, 8);
data_hash[0] = signature_included;
data_hash[1] = (u32)(data_len);
data_hash[2] = (u32)(data_len >> 32);

u32 tmp_data[2];
u32 tmp_cntr = 0;
tmp_data[0] = 0;
tmp_data[1] = 0;


for(u64 ci = 0; ci < data_len; ci++)
 {
 u8 x = fgetc(infile);
 x ^= keystream.get_keystream_byte();
 fputc(x, outfile);

 tmp_data[1] = (tmp_data[1] << 8) | (tmp_data[0] >> 24);
 tmp_data[0] = (tmp_data[0] << 8) | x;
 tmp_cntr++;
 if(tmp_cntr >= 8)
  {
  tmp_cntr = 0;
  hash_block_256(data_hash, tmp_data);

  tmp_data[0] = 0;
  tmp_data[1] = 0;
  }
 }
// Pad with zeroes if required
if(tmp_cntr)
 {
 tmp_cntr = 8 - tmp_cntr;   
 while(tmp_cntr)
  {
  tmp_cntr--;
  tmp_data[1] = (tmp_data[1] << 8) | (tmp_data[0] >> 24);
  tmp_data[0] <<= 8;
  }
 
 hash_block_256(data_hash, tmp_data);
 }

if(signature_included)
 {
 zero(y, signature_length_x32);   
 compose_signature_data(y, data_hash, &secret_data->certificate); 
 ds->encode_signature(&signature, y, secret_data, password, password_length_bytes);        
 fwrite(&signature, sizeof(SIGNATURE), 1, outfile);
 fwrite(&secret_data->certificate, sizeof(SERTIFICATE), 1, outfile);
 }

reset();

return ftell(outfile);
}
//--------------------------------------
// Decrypt file
// Returns length

u64 EL_GAMAL_MODE::decrypt(FILE *outfile, FILE *infile, 
                                              SECRET_DATA const *secret_data, // Recipient key for decryption 
                                              u8 const *password, const u32 password_length_bytes, // Recipient password
                                               int (*verify_certificate)(SERTIFICATE const*)) // Callback for sender verification 
{

    
if(!check_password(&secret_data->encoded_secret_key, password, password_length_bytes)) 
 {
 reset();
 return 0;
 }


u32 fingerprint[fingerprint_length_x32];

fseek(infile, 0, SEEK_SET);
fread(fingerprint, sizeof(u32), fingerprint_length_x32, infile); 

if(0 != multi_compare(fingerprint, secret_data->certificate.fingerprint, fingerprint_length_x32)) // Is this our key?
 {
 reset();
 return 0;
 }


int signature_included = fgetc(infile);
u64 data_len = 0;
fread(&data_len, sizeof(u64), 1, infile);
fread(open_key, sizeof(u32), signature_length_x32, infile);

prepare_decryption_data(secret_data, password, password_length_bytes); 

zero(data_hash, 8);
data_hash[0] = signature_included;
data_hash[1] = (u32)(data_len);
data_hash[2] = (u32)(data_len >> 32);

u32 tmp_data[2];
u32 tmp_cntr = 0;
tmp_data[0] = 0;
tmp_data[1] = 0;


fseek(outfile, 0, SEEK_SET);

keystream.init_key(encryption_key);

for(u64 ci = 0; ci < data_len; ci++)
 {
 u8 x = fgetc(infile);

 tmp_data[1] = (tmp_data[1] << 8) | (tmp_data[0] >> 24);
 tmp_data[0] = (tmp_data[0] << 8) | x;
 tmp_cntr++;
 if(tmp_cntr >= 8)
  {
  tmp_cntr = 0;
  hash_block_256(data_hash, tmp_data);
  tmp_data[0] = 0;
  tmp_data[1] = 0;
  }
 
 x ^= keystream.get_keystream_byte();
 fputc(x, outfile);
 }
// Pad with zeroes if required

if(tmp_cntr)
 {
 tmp_cntr = 8 - tmp_cntr;   
 while(tmp_cntr)
  {
  tmp_cntr--;
  tmp_data[1] = (tmp_data[1] << 8) | (tmp_data[0] >> 24);
  tmp_data[0] <<= 8;
  }
 
 hash_block_256(data_hash, tmp_data);
 }

SIGNATURE signature;
SERTIFICATE certificate;

if(signature_included)
 {
 fread(&signature, sizeof(SIGNATURE), 1, infile);
 fread(&certificate, sizeof(SERTIFICATE), 1, infile);
 
 if(!ds->certificate_looks_ok(&certificate))
  {
  reset();
  return 0;
  }

 ds->decode_signature(y, &signature, &certificate);

 if(0 != multi_compare(y, data_hash, fingerprint_length_x32))
  {
  reset();
  return 0;
  }

 if(verify_certificate != NULL)
  {
  if(!verify_certificate(&certificate))
   {
   reset();
   return 0;
   }
  }
 }


reset();

return data_len;
}

//======================== <END> ============================

