// Encryption/Decryption packets
//
// Entropy is collected as plaintext
// MIC is calculated over ciphertext
// 


#include "serialize.h"
#include "hash_func.h"
#include "random.h"
#include "encrypt_decrypt_packet.h"
#include <string.h>
#include "large_math.h"
#include "idea.h"
#include "xtea.h"


ENCRYPT_DECRYPT_PACKET::ENCRYPT_DECRYPT_PACKET(u32 len_bytes)
{
max_data_len = len_bytes;    
ENCRYPT_DECRYPT_PACKET::reset();
}

ENCRYPT_DECRYPT_PACKET::~ENCRYPT_DECRYPT_PACKET()
{
ENCRYPT_DECRYPT_PACKET::reset();
}
//----------------------------------------
// Initialize all keys from key_data
//
// key data is: 
// mic_iv[2]
// mic_key[4]
// keystream_key[12]
// = total of u32[18] (576 bits)
//
void ENCRYPT_DECRYPT_PACKET::init(u32 const *key_data)
{
copy(mic_iv, key_data, 2);
key_data += 2;

copy(mic_key, key_data, 4);
key_data += 4;

keystream.init_key(key_data); // Initialize PRNG key

reset_stack();

invalid_packets_cntr = 0;
late_packets_cntr = 0;
}
//=============================================================
// Compute Message Identification Code (to append or to verify)
//
u32 ENCRYPT_DECRYPT_PACKET::compute_mic(u8 const *packet)
{
// Calculate hash over encoded data
// Initialize with mic_iv

u32 len_bytes = bytes_to_u32(packet + packet_length_offset);

u32 tmp_hash[2];

tmp_hash[0] = mic_iv[0];
tmp_hash[1] = mic_iv[1];

// Hash with XTEA

Hash_Block_X_64(tmp_hash, packet, len_bytes + 4 + 8); // include counter + len + data

// Encrypt with IDEA

IDEA_Encode(tmp_hash, mic_key);

// Return 32 bit value

return (tmp_hash[0] + tmp_hash[1]);
}
// Destroy all secret data

void ENCRYPT_DECRYPT_PACKET::reset()
{
invalid_packets_cntr = 0;
late_packets_cntr = 0;

cheap_random_number(mic_iv, 2);
cheap_random_number(mic_key, 4);
keystream.reset();
reset_stack();
}

u32 ENCRYPT_DECRYPT_PACKET::encrypt(u8 *dest, u8 const *src, const u32 src_len_bytes)
{
if(!src_len_bytes) return 0;
    
if(src_len_bytes > max_data_len) return 0;


// Put data length into packet 

u32_to_bytes(dest, src_len_bytes);    

// Put counter into packet    

u64 keystream_cntr = keystream.counter();

u64_to_bytes(dest + packet_counter_offset, keystream_cntr);


// Collect entropy

collect_entropy(src, src_len_bytes);


// Encode data by XORing with keystream

for(u32 ci = 0; ci < src_len_bytes; ci++)
 {
 dest[ci + packet_data_offset] = src[ci] ^ keystream.get_keystream_byte();
 }

// Calculate hash over encoded data

u32 mic = compute_mic(dest);

// Append hash

u32_to_bytes(dest + packet_data_offset + src_len_bytes, mic);

return (src_len_bytes + packet_data_offset + 4); // added MIC
}

//======================================================================
// Return 0 = invalid packet (out of sequence, corrupt, or invalid MIC)
//
u32 ENCRYPT_DECRYPT_PACKET::decrypt(u8 *dest, u8 const *src, const u32 src_len_bytes)
{

if(src_len_bytes <= packet_overhead_bytes)
 {
 invalid_packets_cntr++; 
 return 0;
 }

if(src_len_bytes > (packet_overhead_bytes + max_data_len))
 {
 invalid_packets_cntr++; 
 return 0;
 }
   
// Get data length 

u32 len = bytes_to_u32(src + packet_length_offset);

if(!len)
 {
 invalid_packets_cntr++; 
 return 0;
 }

if(len > max_data_len) 
 {
 invalid_packets_cntr++; 
 return 0;
 }

if((len + packet_overhead_bytes) > src_len_bytes) 
 {
 invalid_packets_cntr++; 
 return 0;
 }

// Check MIC

u32 mic = compute_mic(src);

u32 mic_ = bytes_to_u32(src + packet_data_offset + len);

if(mic != mic_) 
 {
 invalid_packets_cntr++; 
 return 0; // Invalid MIC
 }

// Get counter    

u64 cntr = bytes_to_u64(src + packet_counter_offset);


u64 keystream_cntr = keystream.counter();

if(cntr < keystream_cntr) 
 {
 late_packets_cntr++;
 return 0; // Invalid packet
 }

if((cntr - keystream_cntr) > max_skip_counter) 
 {
 invalid_packets_cntr++;   
 return 0; // Too much forward
 }

if(!keystream.counter(cntr)) 
 {
 invalid_packets_cntr++; 
 return 0; // Invalid cntr
 }
// Decode data by XORing with keystream

for(u32 ci = 0; ci < len; ci++)
 {
 dest[ci] = src[ci + packet_data_offset] ^ keystream.get_keystream_byte();
 }

// Collect entropy

collect_entropy(dest, len);

return len;
}

u64 ENCRYPT_DECRYPT_PACKET::total_rx_bytes()
{
return keystream.counter();
}

u64 ENCRYPT_DECRYPT_PACKET::dropped_rx_bytes()
{
return keystream.skip_counter();
}

u64 ENCRYPT_DECRYPT_PACKET::late_packets()
{
return late_packets_cntr;
}

u64 ENCRYPT_DECRYPT_PACKET::invalid_packets()
{
return invalid_packets_cntr;
}
//-----------
// The hash of data is calculated BEFORE stream cipher
// This is done to detect out of sync stream cipher condition
//
u16 ENCRYPT_DECRYPT_PACKET::small_encrypt(u8 *dest, u8 const *src, const u16 src_len_bytes)  // Packet encryption, returns length of encrypted block
{
if(!src_len_bytes) return 0; // must be some data

if(src_len_bytes > small_encrypt_max_data_size) return 0;

// Collect entropy

collect_entropy(src, src_len_bytes);

// Pack counter

u16_to_bytes(dest, (u16)(keystream.counter() & 0xFFFF));

// Pack data

memcpy(dest + 2, src, src_len_bytes);

// Compute MIC

u32 mic = small_compute_mic(src, src_len_bytes);

// Pack MIC

u32_to_bytes(dest + 2 + src_len_bytes, mic);

// Encrypt data + MIC

u16 len = src_len_bytes + small_encrypt_packet_overhead_bytes; 

for(u16 i = 2; i < len; i++) dest[i] ^= keystream.get_keystream_byte();

return len;
}

//-----------------------------------

u16 ENCRYPT_DECRYPT_PACKET::small_decrypt(u8 *dest, u8 const *src, const u16 src_len_bytes)  // Packet decryption, returns length of decrypted block (length -> 0 = invalid block)
{
u16 result = 0;    

if(src_len_bytes <= small_encrypt_packet_overhead_bytes)
 {
 invalid_packets_cntr++;
 return 0;
 }

// Save current state

tmp_keystream_state = keystream;

u16 cntr16_from_packet = bytes_to_u16(src);

// Attempt to decrypt

result = src_len_bytes - small_encrypt_packet_overhead_bytes;

u64 cntr = keystream.counter();
u16 cntr16 = (u16)(cntr & 0xFFFF);

u16 roll_forward = cntr16_from_packet - cntr16;

if(roll_forward > small_encrypt_max_skip_forward_bytes) 
 {
 roll_forward = 0;
 result = 0;
 late_packets_cntr++;
 }



cntr += roll_forward;
 

if(!keystream.counter(cntr)) // This should never happen: overrun of 2^64 boundary
 {
 result = 0;
 }


if(result)
 {
 for(u16 i = 0; i < result; i++) dest[i] = src[i + 2] ^ keystream.get_keystream_byte();

 u32 mic_computed = small_compute_mic(dest, result);

 u8 tmp_mic[sizeof(u32)];
 for(u16 i = 0; i < sizeof(u32); i++) tmp_mic[i] = src[result + 2 + i] ^ keystream.get_keystream_byte();

 u32 mic = bytes_to_u32(tmp_mic);

 if(mic != mic_computed) result = 0;
 }

if(!result) // Restore state
 {
 invalid_packets_cntr++;
 keystream = tmp_keystream_state;
 }
else
 {
 collect_entropy(dest, result);
 }

// Wash saved state

tmp_keystream_state.reset();

return result;
}
     
//=============================================================
// Compute Message Identification Code (to append or to verify)
//
u32 ENCRYPT_DECRYPT_PACKET::small_compute_mic(u8 const *data, const u16 data_len_bytes) // Length excluding mic
{
// Calculate hash over encoded data
// Initialize with mic_iv


u32 tmp_hash[2];

tmp_hash[0] = mic_iv[0];
tmp_hash[1] = mic_iv[1];

// Include length!

tmp_hash[0] ^= data_len_bytes;

// Hash with XTEA

Hash_Block_X_64(tmp_hash, data, data_len_bytes);

// Encrypt with IDEA

IDEA_Encode(tmp_hash, mic_key);

// Return 32 bit value

return (tmp_hash[0] + tmp_hash[1]);
}


//======================== <END> ===================
