// Calculate inverse of G modulo P
// Euclidean algorithm

#include "types.h"
#include "large_math.h"
#include "inverse.h"
#include "random.h"


//
// Returns 1 if inverse exists (p and G are coprime)
//


INVERSE_MODULO::INVERSE_MODULO(u32 length_x32)
{
len = 2*length_x32; 

smul = new SIGNED_MULTI_MUL(len);
sdiv = new SIGNED_MULTI_DIV(len);

quotient = new u32[len];
prod = new u32[2*len];

for(u32 ci = 0; ci < 3; ci++)
 {
 x[ci] = new u32[len];
 y[ci] = new u32[len];
 r[ci] = new u32[len];
 }

}

INVERSE_MODULO::~INVERSE_MODULO()
{
INVERSE_MODULO::reset();

for(u32 ci = 3; ci > 0; ci--)
 {
 delete[] r[ci-1];
 delete[] y[ci-1];
 delete[] x[ci-1];
 }

delete[] prod;
delete[] quotient;
delete sdiv;
delete smul;
}

void INVERSE_MODULO::reset()
{
for(u32 ci = 0; ci < 3; ci++)
 {
 cheap_random_number(x[ci],len);
 cheap_random_number(y[ci],len);
 cheap_random_number(r[ci],len);
 }    

cheap_random_number(quotient,len);

cheap_random_number(prod,2*len);

smul->reset();
sdiv->reset();
}

//--------------------
// Euclidean algorithm
// px + gy = 1 
//
// calculates inverse of g modulo p
// 2 <= g < p
// Returns 1 if inverse exists (p and g are coprime)
//
int INVERSE_MODULO::inverse(u32 *g_, u32 const* g, u32 const* p)
{
// Initialize

int ret_value = 0;
u32 base_idx = 0;
u32 idx[3];

zero(x[0], len);  // x0 = 1
x[0][0] = 1;
zero(x[1], len);  // x1 = 0

zero(y[0], len);  // y0 = 0
zero(y[1], len);  // y1 = 1
y[1][0] = 1;

zero(r[0], len);  // r0 = p
copy(r[0], p, len/2);

zero(r[1], len); // r1 = g
copy(r[1], g, len/2);

// Iteration

for(;;)
 {
 // Calculate indexes (0 - oldest, 2 - newest)

 idx[0] = base_idx;
 idx[1] = (base_idx + 1) % 3;
 idx[2] = (base_idx + 2) % 3;
 base_idx = idx[1];

 sdiv->signed_multi_div(quotient, r[idx[2]], r[idx[0]], r[idx[1]]); // ri = r0/r1

 if(equal_to_zero(r[idx[2]], len))
  {
  if(y[idx[1]][len-1] & 0x80000000) // convert negative modulo to positive
   {  
   multi_add(y[idx[1]],y[idx[1]],p, len/2); 
   }

  copy(g_, y[idx[1]], len/2);
  
  if(equal_to_one(r[idx[1]],len)) ret_value = 1; // GCD = 1 -> Inverse exists
  
  break;
  }
 
 smul->signed_multi_mul(prod, quotient, x[idx[1]]);
 multi_sub(x[idx[2]], x[idx[0]], prod, len);

 smul->signed_multi_mul(prod, quotient, y[idx[1]]);
 multi_sub(y[idx[2]], y[idx[0]], prod, len);
 }

return ret_value;
}

//================== <END> ====================
