// Obfuscation of the data
//
#include "obfuscation.h"
#include "random.h"
#include <string.h>

static const u32 packet_max_size_bytes = 512;

//
// S-box from AES
// (there is no special purpose for using AES S-box here, but anyway)
//
static const u8 s_box[256] = 
{
0x63,0x7c,0x77,0x7b,0xf2,0x6b,0x6f,0xc5,0x30,0x01,0x67,0x2b,0xfe,0xd7,0xab,0x76,
0xca,0x82,0xc9,0x7d,0xfa,0x59,0x47,0xf0,0xad,0xd4,0xa2,0xaf,0x9c,0xa4,0x72,0xc0,
0xb7,0xfd,0x93,0x26,0x36,0x3f,0xf7,0xcc,0x34,0xa5,0xe5,0xf1,0x71,0xd8,0x31,0x15,
0x04,0xc7,0x23,0xc3,0x18,0x96,0x05,0x9a,0x07,0x12,0x80,0xe2,0xeb,0x27,0xb2,0x75,
0x09,0x83,0x2c,0x1a,0x1b,0x6e,0x5a,0xa0,0x52,0x3b,0xd6,0xb3,0x29,0xe3,0x2f,0x84,
0x53,0xd1,0x00,0xed,0x20,0xfc,0xb1,0x5b,0x6a,0xcb,0xbe,0x39,0x4a,0x4c,0x58,0xcf,
0xd0,0xef,0xaa,0xfb,0x43,0x4d,0x33,0x85,0x45,0xf9,0x02,0x7f,0x50,0x3c,0x9f,0xa8,
0x51,0xa3,0x40,0x8f,0x92,0x9d,0x38,0xf5,0xbc,0xb6,0xda,0x21,0x10,0xff,0xf3,0xd2,
0xcd,0x0c,0x13,0xec,0x5f,0x97,0x44,0x17,0xc4,0xa7,0x7e,0x3d,0x64,0x5d,0x19,0x73,
0x60,0x81,0x4f,0xdc,0x22,0x2a,0x90,0x88,0x46,0xee,0xb8,0x14,0xde,0x5e,0x0b,0xdb,
0xe0,0x32,0x3a,0x0a,0x49,0x06,0x24,0x5c,0xc2,0xd3,0xac,0x62,0x91,0x95,0xe4,0x79,
0xe7,0xc8,0x37,0x6d,0x8d,0xd5,0x4e,0xa9,0x6c,0x56,0xf4,0xea,0x65,0x7a,0xae,0x08,
0xba,0x78,0x25,0x2e,0x1c,0xa6,0xb4,0xc6,0xe8,0xdd,0x74,0x1f,0x4b,0xbd,0x8b,0x8a,
0x70,0x3e,0xb5,0x66,0x48,0x03,0xf6,0x0e,0x61,0x35,0x57,0xb9,0x86,0xc1,0x1d,0x9e,
0xe1,0xf8,0x98,0x11,0x69,0xd9,0x8e,0x94,0x9b,0x1e,0x87,0xe9,0xce,0x55,0x28,0xdf,
0x8c,0xa1,0x89,0x0d,0xbf,0xe6,0x42,0x68,0x41,0x99,0x2d,0x0f,0xb0,0x54,0xbb,0x16
};




static inline u8 obfuscation_prng(u32 &state)
{
state ^= 0x259f1ba9; // Constant is M-sequence of length 31 extended by 1
state = (state << 8) - state;
return (u8)(state >> 24);
}

u32 Obfuscate_Packet(u8 *dst, u8 const *src, const u32 src_len_bytes, const u8 packet_padding_enable)
{
if(src_len_bytes > (packet_max_size_bytes - 2*sizeof(u32))) return 0;

u32 nonce;

cheap_random_number(&nonce, 1);

memcpy(dst, &nonce, sizeof(u32));
dst += sizeof(u32);

u8 *ptr = (u8*)(&src_len_bytes);
for(u32 i = 0; i < sizeof(u32); i++) *dst++ = obfuscation_prng(nonce) ^  (*ptr++);

for(u32 i = 0; i < src_len_bytes; i++) *dst++ = obfuscation_prng(nonce) ^  (*src++);

if(!packet_padding_enable) return 2*sizeof(u32) + src_len_bytes;
 

u32 padding_len_bytes = packet_max_size_bytes - src_len_bytes - 2*sizeof(u32);

// To keep padding overhead no more then ~50% on average, padding length = src_length 

if(padding_len_bytes > src_len_bytes) padding_len_bytes = src_len_bytes;


if(padding_len_bytes) 
 {
 cheap_random_number(&nonce, 1); 
 padding_len_bytes = nonce % (padding_len_bytes + 1);
 }

for(u32 i = 0; i < padding_len_bytes; i++) *dst++ = obfuscation_prng(nonce); 

return 2*sizeof(u32) + src_len_bytes + padding_len_bytes;
}

u32 Deobfuscate_Packet(u8 *dst, u8 const *src, const u32 src_len_bytes)
{

if(src_len_bytes <= (2*sizeof(u32))) return 0;

if(src_len_bytes > packet_max_size_bytes) return 0;

u32 nonce, dst_len_bytes;

memcpy(&nonce, src, sizeof(u32));
src += sizeof(u32);

u8 *ptr = (u8*)(&dst_len_bytes);
for(u32 i = 0; i < sizeof(u32); i++) *ptr++ = obfuscation_prng(nonce) ^  (*src++);

if(dst_len_bytes > (src_len_bytes - 2*sizeof(u32))) return 0;

for(u32 i = 0; i < dst_len_bytes; i++) *dst++ = obfuscation_prng(nonce) ^ (*src++); 

return dst_len_bytes;
}
//----------------
// Obfuscation with minimal overhead (+1 byte)
// Padding length is 0...255 bytes
//
static const u32 small_obfuscation_rounds = 4; // ~2^32 variants


u16 Small_Obfuscate_Packet(u8 *dst, u8 const *src, const u16 src_len_bytes, const u8 packet_padding_enable)
{
u32 nonce = 0;

if(!src_len_bytes) return 0;

if(src_len_bytes > (packet_max_size_bytes - 1)) return 0;

memcpy(dst, src, src_len_bytes);

u16 dst_len = src_len_bytes;

u16 padding_len_bytes = packet_max_size_bytes - src_len_bytes - 1;

if(!packet_padding_enable) padding_len_bytes = 0; 

 // No more then 255 bytes

if(padding_len_bytes > 255) padding_len_bytes = 255; 

 // To keep padding overhead no more then ~50% on average, padding length = src_length 

if(padding_len_bytes > src_len_bytes) padding_len_bytes = src_len_bytes;

if(padding_len_bytes) 
 {
 cheap_random_number(&nonce, 1); 
 padding_len_bytes = nonce % (padding_len_bytes + 1);
 }

for(u16 i = 0; i < padding_len_bytes; i++) dst[dst_len++] = obfuscation_prng(nonce); 

dst[dst_len] = (u8)padding_len_bytes;
dst_len++;


for(u8 i = 0; i < small_obfuscation_rounds; i++)
 {
 for(u16 j = 1; j < dst_len; j++) dst[j] ^= s_box[dst[j-1]];
 
 dst[0] ^= s_box[(dst[dst_len - 1] ^ i)];
 }

return dst_len;
}


u16 Small_Deobfuscate_Packet(u8 *dst, u8 const *src, const u16 src_len_bytes)
{
if(src_len_bytes <= 1) return 0;

if(src_len_bytes > packet_max_size_bytes) return 0;

memcpy(dst, src, src_len_bytes);


u8 i = small_obfuscation_rounds;

while(i--)
 {
 dst[0] ^= s_box[(dst[src_len_bytes - 1] ^ i)]; 
    
 for(u16 j = src_len_bytes - 1; j > 0; j--) dst[j] ^= s_box[dst[j-1]];
 }

u16 padding_len_bytes = dst[src_len_bytes - 1];

if(src_len_bytes <= (padding_len_bytes+1)) return 0;

u16 dst_len_bytes = src_len_bytes - padding_len_bytes - 1;

return dst_len_bytes;
}

