// Random numbers 

#include "types.h"
#include "large_math.h"
#include "hash_func.h"
#include "xtea.h"
#include "idea.h"
#include "random.h"
#include "system_random.h"
#include <time.h>
#include "serialize.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <tchar.h>
#include "system_critical_section.h"


static SYSTEM_CRITICAL_SECTION main_rng_critical_section; // Must be static


// The main seed is 1 KB size

static const u32 seed_size_u32 = 256;


static u32 default_seed[seed_size_u32]; // main seed
static u32 entropy_buffer[seed_size_u32]; // collect entropy
static u32 entropy_N = 0;  // Increments at every shuffle
static u32 entropy_byte_cnt = 0; // Increments at every byte


static u32 cheap_seed[2]; // To generate junk




//--------------------------------------
// Initialize seed

#ifndef _UNICODE

static char seed_path[MAX_PATH + 1] = "seed.bin";

void initialize_seed(const char *path)
{
if(path != NULL)
 {
 strncpy(seed_path, path, sizeof(seed_path));
 }

u8 buff[seed_size_u32*sizeof(u32)];

FILE *file = fopen(seed_path,"rb");

if(file != NULL)
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

file = fopen(seed_path,"wb");
fwrite(default_seed, seed_size_u32*sizeof(u32), 1, file);
fclose(file);

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}
#else

static TCHAR seed_path[MAX_PATH + 1] = _T("seed.bin");

void initialize_seed(const TCHAR *path)
{
if(path != NULL)
 {
 _tcsncpy(seed_path, path, MAX_PATH + 1);
 }

u8 buff[seed_size_u32*sizeof(u32)];

FILE *file;

if(!_tfopen_s(&file,seed_path,_T("rb")))
 {
 fread(buff, sizeof(buff), 1, file); 
 fclose(file);
 
 collect_entropy(buff, sizeof(buff));
 }

// Wash seed

collect_entropy();

if(!_tfopen_s(&file,seed_path,_T("wb")))
 {
 fwrite(default_seed, seed_size_u32*sizeof(u32), 1, file);
 fclose(file);
 }

// Initialize cheap seed
 
time_t t;
time(&t);
Hash_Block_X_64(cheap_seed, (u8 *)&t, sizeof(time_t));

// Cleanup

cheap_random_number(buff, seed_size_u32); 
reset_stack();
}

#endif


//--------------------
// Optimized bijection
// (differential strength ~ 3 bits per round)
//
inline static u32 mix32(u32 x)
{
x = (x<<4) - x;           // multiply by 15
x ^= x >> 10;             // xor
x = (x << 7) | (x >> 25); // cyclic shift left by 7
return x;
}

// UFN

inline static void mix32_round(u32 *ptr, u32 &N, const u32 len)
{

for(u32 i = 1; i < len; i++)
 {
 ptr[i] += mix32(ptr[i-1] ^ N);
 N++;
 }

ptr[0] += mix32(ptr[len-1] ^ N);
N++;
}

//------------------------------------
// Uninvertible keyless function
//

static void mix_entropy(u32 *x0, u32 *x1, u32 &N)
{
u32 lock_bar[2*seed_size_u32];    

const u32 entropy_hash_rounds = 20; // conservative, ~13 should be enough

copy(lock_bar, x0, seed_size_u32);
copy(lock_bar + seed_size_u32, x1, seed_size_u32);


// Feistel

for(u32 i = 0; i < entropy_hash_rounds; i++)
 {
 xor_data(lock_bar, lock_bar + seed_size_u32, seed_size_u32);
 mix32_round(lock_bar, N, seed_size_u32);

 xor_data(lock_bar + seed_size_u32, lock_bar, seed_size_u32);
 mix32_round(lock_bar + seed_size_u32, N, seed_size_u32);    
 }

// Lock data

xor_data(x0, lock_bar, seed_size_u32);
xor_data(x1, lock_bar + seed_size_u32, seed_size_u32);


cheap_random_number(lock_bar, 2*seed_size_u32); 
reset_stack();
}
//----------------------------
// Collect entropy
// When 1K is collected, seed and buffer are shuffled
//
//
void collect_entropy(void const *data, u32 data_size_bytes)
{


if((data == NULL) || (data_size_bytes == 0))
 {
 collect_entropy();
 return;
 }

main_rng_critical_section.EnterCriticalSection();

u8 *dst = (u8*)entropy_buffer;
u8 *src = (u8*)data;

 
while(data_size_bytes--)
 {
 dst[entropy_byte_cnt] ^= *src++;
 entropy_byte_cnt++;

 if(entropy_byte_cnt >= sizeof(entropy_buffer))
  {   
  entropy_byte_cnt = 0;
  mix_entropy(default_seed, entropy_buffer, entropy_N);
  }
 } // while


main_rng_critical_section.LeaveCriticalSection();
}


//-----------------------
// System provided entropy
// Note: system call can be easily captured
// So it should not be called in main RNG
//
void collect_entropy()
{
const u32 buff_size_bytes = 1024; // Draw 1K
u8 buff[buff_size_bytes];    

Get_System_Provided_Entropy(buff, buff_size_bytes);

collect_entropy(buff, buff_size_bytes);
cheap_random_number(buff, sizeof(buff));

time_t t;
time(&t);

collect_entropy(&t, sizeof(time_t));
cheap_random_number((u8 *)&t, sizeof(time_t));



reset_stack();
}

//----
// u32[2] for legacy compatibility
//
void collect_entropy(u32 const* data)
{
collect_entropy(data, 2*sizeof(u32));
}

/*
//-----------------------------------------------------
// Good but too slow as it shuffles entropy every time

void collect_entropy(void const *data, const u32 data_size_bytes)
{

if((data == NULL) || (data_size_bytes == 0))
 {
 collect_entropy();
 return;
 }

u32 lock[seed_size_u32];
u32 k[seed_size_u32];
const u32 seed_size_bytes = seed_size_u32*sizeof(u32);

main_rng_critical_section.EnterCriticalSection();

copy(lock, default_seed, seed_size_u32);

main_rng_critical_section.LeaveCriticalSection();


const u32 fixed_rounds = 20;

u32 rounds = fixed_rounds;

if(data_size_bytes > seed_size_bytes) 
 {
 u32 tmp = data_size_bytes*fixed_rounds;   
 rounds = tmp/seed_size_bytes;
 if(tmp%seed_size_bytes) rounds++;
 }

u32 data_idx = 0;
u32 N = 0; // tweak value

while(rounds)
 {
 rounds--;
 
 // Initialize key

 u32 key_idx = 0;
 
 do {
   
    u32 data_bytes_remaining = data_size_bytes - data_idx;
    u32 key_bytes_to_fill  = seed_size_bytes - key_idx;

    u32 data_bytes_to_copy;
 
    if(data_bytes_remaining <= key_bytes_to_fill) data_bytes_to_copy = data_bytes_remaining;
    else data_bytes_to_copy = key_bytes_to_fill;
    
    memcpy(((u8*)k) + key_idx, ((u8*)data) + data_idx, data_bytes_to_copy);

    data_idx += data_bytes_to_copy;
    if(data_idx >= data_size_bytes) data_idx = 0;

    key_idx  += data_bytes_to_copy; 
    
    } while(key_idx < seed_size_bytes);

 // Round

 for(u32 i = 1; i < seed_size_u32; i++) 
  {
  lock[i] += mix32((lock[i-1] ^ k[i]) + N);
  N++;
  }

 lock[0] += mix32((lock[seed_size_u32 - 1] ^ k[0]) + N);
 N++;
 } // rounds

// Lock up

main_rng_critical_section.EnterCriticalSection();

xor_data(default_seed, lock, seed_size_u32);

main_rng_critical_section.LeaveCriticalSection();

// Cleanup

cheap_random_number(lock, seed_size_u32);
cheap_random_number(k, seed_size_u32);
reset_stack();
}
*/


void reset_stack(void)
{
const u32 dummy_size = 256;    
u32 dummy[dummy_size];
cheap_random_number(dummy, dummy_size);
}

//----------------------------------------------------
// No need to protect cheap rng with critical sections

u32 cheap_random_number()
{
return lcg64(cheap_seed);
}
 
void cheap_random_number(u32 *x, const u32 length_x32)
{
for(u32 i = 0; i < length_x32; i++) x[i] = lcg64(cheap_seed);
}

void cheap_random_number(u8 *x, const u32 size_of_x_bytes)
{
for(u32 i = 0; i < size_of_x_bytes; i++) x[i] = (u8)lcg64(cheap_seed);
}

//---------------------------
// Simple 64-bit LCG
// Do NOT change since it is used in keystream generation and hash functions
//
u32 lcg64(u32 *state)
{
u64 x = state[0] | (((u64)state[1])<<32);
x = (x << 5) + x + 0x71E86A1CD2D0E1DBllu;
state[0] = (u32)x;
state[1] = (u32)(x>>32);
return (state[0] ^ state[1]);
}


//--------------------------------------------
// Generation of strong random numbers
//

void random_number(u32 *x, const u32 length_x32)
{
u32 tmp[2];

main_rng_critical_section.EnterCriticalSection();

//---------------------------------------
// Prehash
// Do not call system entropy here
// Because attacker can capture system call
//
mix_entropy(default_seed, entropy_buffer, entropy_N);    
  

// Generate sequence

u32 cnt = length_x32;

while(cnt >= 2)
 {
 cnt -= 2; 
 universal_prng(x, default_seed, seed_size_u32);
 x += 2;
 }

if(cnt)
 {
 universal_prng(tmp, default_seed, seed_size_u32);
 *x = tmp[0];
 }

//---------------------------------------
// Posthash
// Do not call system entropy here
// Because attacker can capture system call
//
mix_entropy(default_seed, entropy_buffer, entropy_N);    

main_rng_critical_section.LeaveCriticalSection();

//  Cleanup

cheap_random_number(tmp, 2);
reset_stack();
}

//----------------------
//  x ^= C
//  x = ((x << 32) - x);
// Full 128 bit period
//
u32 nlcg128(u32 *state)
{
const u32 C = 0xAA551EE1; // Any odd number

u32 tmp32 = state[0];
tmp32 ^= C;
s64 tmp64 = 0;
tmp64 -= tmp32;
state[0] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[1];
tmp64 -= tmp32;
state[1] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[2];
tmp64 -= tmp32;
state[2] = (u32)tmp64;
tmp64 >>= 32; // Arithmetic shift
tmp64 += tmp32;
tmp32 = state[3];
tmp64 -= tmp32;
state[3] = (u32)tmp64;

return (state[3] ^ state[2]);
}

//------------------------------------------------
// State is [k0].....k[11]
// 
// state u32[12] = 384 bit state
// state[0..3] is 128 bit NLCG
// state[4..11] is 256 bit hash
//
// Generates u32[2] pseudorandom number 
//
// Irreversible function 256 bit strong
//
void keystream_generator(u32 *output, u32 *state)
{
const u32 rounds = keystream_xxtea_rounds;

u32 lock_bar[8];

//---- State modification ---

// NLCG

nlcg128(state);

// Hash function

copy(lock_bar, state + 4, 8);

XXTEA_Encode(state + 4, 8, state, rounds);

xor_data(state + 4, lock_bar, 8);

cheap_random_number(lock_bar, 8); // wash previous state

// Output encoding 

output[0] = state[2];
output[1] = state[3];

IDEA_Encode(output, state + 4);
}

//===============
// Initialize keystream state u32[12] with key u32[12]
//
void keystream_generator_init(u32 *state, u32 const *key)
{
// Use 384 bits of key as is

copy(state, key, 12);

reset_stack();
}
//---------------------------------------------

KEYSTREAM_GENERATOR::KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset();
}

KEYSTREAM_GENERATOR::~KEYSTREAM_GENERATOR()
{
KEYSTREAM_GENERATOR::reset(); 
}

void KEYSTREAM_GENERATOR::reset()
{
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
memset(byte_buffer,0, sizeof(byte_buffer));
cheap_random_number(state, 12);
reset_stack();
}

void KEYSTREAM_GENERATOR::init_key(u32 const* key)
{
keystream_generator_init(state, key);
cntr = 0;
skip_cntr = 0;
byte_buffer_idx = 0;
}

u8 KEYSTREAM_GENERATOR::get_keystream_byte()
{

if(!byte_buffer_idx) // time to generate next 8 bytes
 {
 u32 tmp_buf[2];   
 keystream_generator(tmp_buf, state);
 
 for(u32 ci = 0; ci < 4; ci++)
  {
  byte_buffer[ci]   = (u8)(tmp_buf[0] & 0xFF);
  tmp_buf[0] >>= 8;

  byte_buffer[ci+4] = (u8)(tmp_buf[1] & 0xFF);  
  tmp_buf[1] >>= 8;
  }
 }

u8 ret_value = byte_buffer[byte_buffer_idx & 0x07]; // Could be = 8 in fast fwd mode
byte_buffer_idx = (byte_buffer_idx + 1) & 0x07;

cntr++;


return ret_value;
}
//--------------------------------------
// Update state without generating keystream data
// (as if 8 bytes per iteration)
//
void KEYSTREAM_GENERATOR::fast_forward(u64 fwd_bytes)
{
if(!fwd_bytes) return; // Nothing to do

cntr += fwd_bytes;

u32 bytes_in_buffer = 0x07 & (8 - byte_buffer_idx); // 0...7

if(fwd_bytes >= bytes_in_buffer)
 {
 u64 diff_bytes = fwd_bytes - bytes_in_buffer;
 u64 fwd_cycles = 1 + (diff_bytes >> 3);

 byte_buffer_idx = (u32)(diff_bytes & 0x07);
 if(!byte_buffer_idx) byte_buffer_idx = 8;

 const u32 rounds = keystream_xxtea_rounds;

 u32 lock_bar[8];    
 u32 tmp_buf[2];

// Roll state forward by number of 8 byte blocks

 while(fwd_cycles)
  {
  fwd_cycles--;

  // NLCG

  nlcg128(state);

  // Hash function

  copy(lock_bar, state + 4, 8);

  XXTEA_Encode(state + 4, 8, state, rounds);

  xor_data(state + 4, lock_bar, 8);
  } // while


 // Generate buffer at last cycle

 tmp_buf[0] = state[2];
 tmp_buf[1] = state[3];

 IDEA_Encode(tmp_buf, state + 4);

 for(u32 ci = 0; ci < 4; ci++)
  {
  byte_buffer[ci]   = (u8)(tmp_buf[0] & 0xFF);
  tmp_buf[0] >>= 8;

  byte_buffer[ci+4] = (u8)(tmp_buf[1] & 0xFF);  
  tmp_buf[1] >>= 8;
  }

 // Cleanup

 cheap_random_number(lock_bar, 8);
 cheap_random_number(tmp_buf, 2);
 reset_stack();
 }
else // No need to update state, just move pointer
 {
 byte_buffer_idx += (u32)fwd_bytes;
 }

}

//------------------------------------------------


int KEYSTREAM_GENERATOR::counter(u64 x)
{
if(x < cntr) return 0; // Can't go back in time


/* Slow

while(x > cntr)
 {
 skip_cntr++;
 get_keystream_byte(); // Roll forward
 }
*/

if(x > cntr)
 {
 u64 diff = x - cntr;

 skip_cntr += diff;

 fast_forward(diff);
 }

return 1;
}

u64 KEYSTREAM_GENERATOR::counter()
{
return cntr;
}

u64 KEYSTREAM_GENERATOR::skip_counter()
{
return skip_cntr;
}


//------------------------------------------------------------
// Generate ~256 bit strong sequence guaranteed > 2^131 bytes length
// using state (state_size_x32 >= 8)
// Output: data[2]
//
void universal_prng(u32 *data, u32 *state, const u32 state_size_x32)
{
const u32 xxtea_rounds = 20; // conservative (~10 should be enough)    

nlcg128(state);

XXTEA_Encode(state + 4, state_size_x32 - 4, state, xxtea_rounds);

data[0] = state[0];
data[1] = state[1];

IDEA_Encode(data, state + 4);
}

//==================== <END> ========================

